#include "cac2obj.h"
#include "cexcept.h"
#include "cpmt.h"

//---------------------------------------------------------------------------------------
CAc2Obj::CAc2Obj()
{
   mF = NULL;
   mFMat = NULL;
   mV = NULL;
   mVSorted = NULL;
   mMatList = NULL;
   mVCount = 0;
   mMatListCount = 0;
}
//---------------------------------------------------------------------------------------
CAc2Obj::~CAc2Obj()
{
   ClearVerts();
   ClearFiles();
}
//---------------------------------------------------------------------------------------
void CAc2Obj::ClearFiles()
{
   if( mF != NULL )
   {
      fclose( mF );
      mF = NULL;
   }
   if( mFMat != NULL )
   {
      fclose( mFMat );
      mFMat = NULL;
   }
}
//---------------------------------------------------------------------------------------
void CAc2Obj::ClearVerts()
{
   if( mV != NULL )
   {
      delete [] mV;
      mV = NULL;
   }
   if( mVSorted != NULL )
   {
      delete [] mVSorted;
      mVSorted = NULL;
   }
   mVCount = 0;
   if( mMatList != NULL )
   {
      delete [] mMatList;
      mMatList = NULL;
   }
   mMatListCount = 0;
}
//---------------------------------------------------------------------------------------
// Vertex sorter
void CAc2Obj::VVertsort( CVertSortItem *vlist, int left, int right )
{
   int i = left;
   int j = right;
   CVertSortItem x = vlist[( left + right ) / 2];
   CVertSortItem y;
   do
   {
      while( vlist[i].mV < x.mV && i < right )
      {
         ++i;
      }
      while( x.mV < vlist[j].mV && j > left )
      {
         --j;
      }
      if( i <= j )
      {
         y = vlist[i];
         vlist[i] = vlist[j];
         vlist[j] = y;
         ++i;
         --j;
      }
   } while( i <= j );
   if( left < j )
   {
      VVertsort( vlist, left, j );
   }
   if( i < right )
   {
      VVertsort( vlist, i, right );
   }
}
//---------------------------------------------------------------------------------------
// Find a vertex within a sorted list
int CAc2Obj::VVertfind( CVertSortItem *vlist, int count, Vertex *v )
{
   int low = 0;
   int high = count - 1;
   int mid;
   while( low <= high )
   {
      mid = ( low + high ) / 2;
      if( v < vlist[mid].mV )
      {
         high = mid - 1;
      }
      else if( v > vlist[mid].mV )
      {
         low = mid + 1;
      }
      else
      {
         return mid;
      }
   }
   return -1;
}
//---------------------------------------------------------------------------------------
void CAc2Obj::VValidate( CVertSortItem *vsorted, ACObject *o, unsigned int &numFound, bool &hasOrphans )
{
   CPmt *pmt = CPmt::GetSingleton();
   for( List *l = ac_object_get_vertexlist( o ); l != NULL; l = l->next )
   {
      if( VVertfind( vsorted, pmt->mVCount, ( Vertex * )( l->data )) == -1 )
      {
         hasOrphans = true;
         return;
      }
      else
      {
         ++numFound;
      }
   }
   for( List *l = ac_object_get_childrenlist( o ); l != NULL; l = l->next )
   {
      VValidate( vsorted, ( ACObject * )( l->data ), numFound, hasOrphans );
      if( hasOrphans )
      {
         return;
      }
   }
}
//---------------------------------------------------------------------------------------
void CAc2Obj::VCollectMats( ACObject *o, CMatItem *matlist, unsigned int matlistcount )
{
   if( matlistcount > 0 )
   {
      for( List *l = ac_object_get_surfacelist( o ); l != NULL; l = l->next )
      {
         unsigned int col = ( unsigned int )((( Surface * )( l->data ))->col );
         if( col < matlistcount && matlist[col].mEntity == NULL )
         {
            matlist[col].mEntity = ( ACEntity * )( ac_palette_get_material( col ));
            char *matname;
            ac_entity_get_string_value( matlist[col].mEntity, "name", &matname );
            if( matname == NULL || matname[0] == '\0' )
            {
               matlist[col].mMatName = new char[16];
               sprintf( matlist[col].mMatName, "mat_%d", col );
            }
            else
            {
               matlist[col].mMatName = new char[strlen( matname ) + 1];
               strcpy( matlist[col].mMatName, matname );
            }
         }
      }
   }
   for( List *l = ac_object_get_childrenlist( o ); l != NULL; l = l->next )
   {
      VCollectMats(( ACObject * )( l->data ), matlist, matlistcount );
   }
}
//---------------------------------------------------------------------------------------
void CAc2Obj::IdentifyAcVertices( unsigned int &numFound, bool &hasOrphans )
{
   ClearVerts();
   // Identify the vertices we have to work with
   CPmt *pmt = CPmt::GetSingleton();
   CVertSortItem *vsorted = new CVertSortItem[pmt->mVCount];
   for( unsigned int ix = 0; ix < pmt->mVCount; ++ix )
   {
      vsorted[ix].mV = pmt->mV[ix];
   }
   VVertsort( vsorted, 0, pmt->mVCount - 1 );
   VValidate( vsorted, ac_get_world(), numFound, hasOrphans );
   // No sense hanging out if we've found orphans
   if( hasOrphans )
   {
      delete [] vsorted;
      return;
   }
   // Get the number of ac3d materials and allocate an array of that size
   int i = ac_get_num_palette();
   if( i > 0 )
   {
      mMatListCount = i;
      mMatList = new CMatItem[mMatListCount];
   }
   // Assign the vertices
   if( numFound == pmt->mVCount )
   {
      mV = new Vertex*[numFound];
      // Copy original vertices
      memcpy( mV, pmt->mV, sizeof( Vertex * ) * numFound );
      mVCount = numFound;
      // Copy "new" verts and collect materials
      VCollectMats( ac_get_world(), mMatList, mMatListCount );
      // Add all vertices to sorted list
      mVSorted = new CVertSortItem[mVCount];
      for( unsigned int ix = 0; ix < mVCount; ++ix )
      {
         mVSorted[ix].mOrder = ix + 1;
         mVSorted[ix].mV = mV[ix];
      }
      VVertsort( mVSorted, 0, mVCount - 1 );
   }
   // Delete temp array
   delete [] vsorted;
}
//---------------------------------------------------------------------------------------
void CAc2Obj::WriteFaceData( ACObject *o, unsigned int &lastmat, bool &lastsmooth )
{
   for( List *l = ac_object_get_surfacelist( o ); l != NULL; l = l->next )
   {
      Surface *s = ( Surface * )( l->data );
      // Print material
      if( lastmat != s->col )
      {
         lastmat = s->col;
         fprintf( mF, "usemtl %s\n", mMatList[lastmat].mMatName );
      }
      // Print smooth group
      bool issmooth = ( surface_get_shading( s ) == TRUE );
      if( issmooth != lastsmooth )
      {
         lastsmooth = issmooth;
         fprintf( mF, "s %s\n", issmooth ? "1" : "off" );
      }
      // Print vertex refs
      fprintf( mF, "f" );
      for( List *vl = s->vertlist; vl != NULL; vl = vl->next )
      {
         SVertex *sv = ( SVertex * )( vl->data );
         int ref = VVertfind( mVSorted, mVCount, sv->v );
         if( ref < 0 )
         {
            printf( "Bad vert address %d\n", sv->v );
         }
         else
         {
            fprintf( mF, " %d", mVSorted[ref].mOrder );
         }
      }
      fprintf( mF, "\n" );
   }
   for( List *l = ac_object_get_childrenlist( o ); l != NULL; l = l->next )
   {
      WriteFaceData(( ACObject * )( l->data ), lastmat, lastsmooth );
   }
}
//---------------------------------------------------------------------------------------
void CAc2Obj::SaveToObjFile( char *fname, char *mname, bool vertsonly )
{

   ClearFiles();

   // Create the output file
   mF = fopen( fname, "wt" );
   if( mF == NULL )
   {
      throw CExcept( "error opening file %s", fname );
   }

   if( !vertsonly )
   {

      // Create the material file
      mFMat = fopen( mname, "wt" );
      if( mFMat == NULL )
      {
         throw CExcept( "error opening material file %s", mname );
      }

      // Write the materials
      ACrgb rgb;
      float f;
      for( unsigned int ix = 0; ix < mMatListCount; ++ix )
      {
         ACEntity *e = mMatList[ix].mEntity;
         if( e == NULL )
         {
            continue;
         }
         fprintf( mFMat, "newmtl %s\n", mMatList[ix].mMatName );
         ac_entity_get_float_value( e, "shininess", &f );
         fprintf( mFMat, "Ns %g\n", f );
         ac_entity_get_rgb_value( e, "ambient", &rgb );
         fprintf( mFMat, "Ka %g %g %g\n", rgb.r, rgb.g, rgb.b );
         ac_entity_get_rgb_value( e, "diffuse", &rgb );
         fprintf( mFMat, "Kd %g %g %g\n", rgb.r, rgb.g, rgb.b );
         ac_entity_get_rgb_value( e, "specular", &rgb );
         fprintf( mFMat, "Ks %g %g %g\n", rgb.r, rgb.g, rgb.b );
         ac_entity_get_float_value( e, "transparency", &f );
         f = max( 0.0f, 1.0f - f );
         fprintf( mFMat, "Tr %g\n", f );
      }

      fclose( mFMat );
      mFMat = NULL;

   }

   // Write the ordered vertex data
   for( unsigned int ix = 0; ix < mVCount; ++ix )
   {
      Vertex *v = mV[ix];
      fprintf( mF, "v %g %g %g\n", v->x, v->y, v->z );
   }

   if( !vertsonly )
   {

      char *p = strrchr( mname, '/' );
      if( p == NULL )
      {
         p = mname;
      }
      else
      {
         ++p;
      }
      fprintf( mF, "mtllib %s\n", p );
      bool lastsmooth = false;
      unsigned lastmat = mMatListCount;
      WriteFaceData( ac_get_world(), lastmat, lastsmooth );
   }

   fclose( mF );
   mF = NULL;

}
//---------------------------------------------------------------------------------------
