// TODO: Remove for Mac compiles
#define WINDOWS
#include "ac_plugin.h"

#include "cobj.h"
#include <stdlib.h>
#include "cstrlist.h"

//---------------------------------------------------------------------------------------
CObj::CObj()
{
   mV = NULL;
   mVt = NULL;
   mF = NULL;
   mM = NULL;
   mVCount = 0;
   mVtCount = 0;
   mFCount = 0;
   mMCount = 0;
}
//---------------------------------------------------------------------------------------
CObj::~CObj()
{
   Clear();
}
//---------------------------------------------------------------------------------------
void CObj::Clear()
{
   if( mV != NULL )
   {
      delete [] mV;
      mV = NULL;
   }
   if( mVt != NULL )
   {
      delete [] mVt;
      mVt = NULL;
   }
   if( mF != NULL )
   {
      delete [] mF;
      mF = NULL;
   }
   if( mM != NULL )
   {
      delete [] mF;
      mM = NULL;
   }
   mVCount = 0;
   mVtCount = 0;
   mFCount = 0;
   mMCount = 0;
}
//---------------------------------------------------------------------------------------
static void cobj_CleanupStrList( CStrList &strlist )
{
   // Replace tabs with spaces
   // Remove multiple spaces
   // Left-trim string
   for( CStrNode *sn = strlist.mHead; sn != NULL; sn = sn->mNext )
   {
      char *pdest = sn->mText;
      for( char *p = sn->mText; *p != '\0'; ++p )
      {
         if( *p == '\t' )
         {
            *p = ' ';
         }
         // Skip if this puts the space as the first character or if this puts in a duplicate space
         if( *p == ' ' && ( pdest == sn->mText || *( pdest - 1 ) == ' ' ))
         {
            continue;
         }
         //*pdest = tolower( *p );
         // Better not lowercase - need to maintain case-sensitivity
         *pdest = *p;
         ++pdest;
      }
      *pdest = '\0';
      // Right-trim string
      for( ;; )
      {
         p = strrchr( sn->mText, ' ' );
         if( p == NULL || (( p - sn->mText ) < ( strlen( sn->mText ) - 1 )))
         {
            break;
         }
         *p = '\0';
      }
   }
}
//---------------------------------------------------------------------------------------
int CObj::FindMaterial( char *matname )
{
   for( int ix = 0; ix < mMCount; ++ix )
   {
      if( strcmpi( mM[ix].mName, matname ) == 0 )
      {
         return ix;
      }
   }
   return -1;
}
//---------------------------------------------------------------------------------------
void CObj::LoadFromFile( char *fname )
{

   Clear();

   // Load the OBJ file into a string list
   CStrList strlist;
   debugf( "obj : loading file into string list\n" );
   strlist.LoadFromFile( fname, COBJ_MAXSTRLEN );
   debugf( "obj : cleaning up string list\n" );
   cobj_CleanupStrList( strlist );

   // Get the path to this file, including final slash
   // NOTE: Assumes filenames are delimited by forward slashes
   char pathname[COBJ_MAXSTRLEN];
   strcpy( pathname, fname );
   char *p = strrchr( pathname, '/' );
   if( p != NULL )
   {
      ++p;
      *p = '\0';
   }
   else
   {
      pathname[0] = '\0';
   }

   // Create a material list
   CStrList matlist;

   // Count the vertices and faces
   // Get the material file(s)
   int v = 0;
   int vt = 0;
   int f = 0;
   char *next;
   char token[COBJ_MAXSTRLEN];
   char token2[COBJ_MAXSTRLEN];
   int width;
   for( CStrNode *sn = strlist.mHead; sn != NULL; sn = sn->mNext )
   {
      if( strlen( sn->mText ) == 0 )
      {
         continue;
      }
      next = sn->mText;
      sscanf( next, "%s%n", token, &width );
      next = next + width;
      if( strcmp( token, "v" ) == 0 )
      {
         ++v;
      }
      else if( strcmp( token, "f" ) == 0 )
      {
         ++f;
      }
      else if( strcmp( token, "vt" ) == 0 )
      {
         ++vt;
      }
      else if( strcmp( token, "mtllib" ) == 0 )
      {
         while( *next == ' ' )
         {
            ++next;
         }
         sprintf( token2, "%s%s", pathname, next );
         try
         {
            matlist.LoadFromFile( token2, COBJ_MAXSTRLEN );
         }
         catch( CExcept e )
         {
            printf( "obj : error loading material file %s\n%s\n", token2, e.text );
         }
         catch( ... )
         {
            printf( "obj : error loading material file %s\n", token2 );
         }
      }
   }

   // Cleanup the material list
   cobj_CleanupStrList( matlist );
   // Count the materials
   int m = 0;
   for( CStrNode *sn = matlist.mHead; sn != NULL; sn = sn->mNext )
   {
      if( strlen( sn->mText ) == 0 )
      {
         continue;
      }
      next = sn->mText;
      sscanf( next, "%s%n", token, &width );
      next = next + width;
      if( strcmp( token, "newmtl" ) == 0 )
      {
         ++m;
      }
   }
   // Load the materials
   mM = new CObjMat[m];
   mMCount = m;
   m = 0;
   CObjMat *currmat = NULL;
   for( CStrNode *sn = matlist.mHead; sn != NULL; sn = sn->mNext )
   {
      if( strlen( sn->mText ) == 0 )
      {
         continue;
      }
      next = sn->mText;
      sscanf( next, "%s%n", token, &width );
      next = next + width;
      if( strcmp( token, "newmtl" ) == 0 )
      {
         currmat = &( mM[m++] );
         while( *next == ' ' )
         {
            ++next;
         }
         currmat->SetName( next );
         continue;
      }
      // Skip if we've no material yet
      if( currmat == NULL )
      {
         continue;
      }
      // Read attribute
      if( strcmp( token, "Ka" ) == 0 )
      {
         sscanf( next, "%e %e %e", &( currmat->mAmbient[0] ), &( currmat->mAmbient[1] ), &( currmat->mAmbient[2] ));
      }
      else if( strcmp( token, "Kd" ) == 0 )
      {
         sscanf( next, "%e %e %e", &( currmat->mDiffuse[0] ), &( currmat->mDiffuse[1] ), &( currmat->mDiffuse[2] ));
      }
      else if( strcmp( token, "Ks" ) == 0 )
      {
         sscanf( next, "%e %e %e", &( currmat->mSpecular[0] ), &( currmat->mSpecular[1] ), &( currmat->mSpecular[2] ));
      }
      else if( strcmp( token, "Tr" ) == 0 ||
               strcmp( token, "d" ) == 0 )
      {
         sscanf( next, "%e", &( currmat->mTransparency ));
      }
      else if( strcmp( token, "Ns" ) == 0 )
      {
         sscanf( next, "%e", &( currmat->mShininess ));
      }
      else if( strcmp( token, "map_Ka" ) == 0 )
      {
         currmat->SetTexture( next );
      }
   }

   // Bail if not enough vertices or faces were found
   if( v < 3 || f == 0 )
   {
      throw CExcept( "error loading obj file; %d vertices and %d faces were found; need at least 3 vertices and 1 face", v, f );
   }

   // Allocate arrays
   mV = new CObjVec3[v];
   mVt = new CObjVec2[vt];
   mF = new CObjFace[f];
   mVCount = v;
   mVtCount = vt;
   mFCount = f;

   // Read data
   v = 0;
   vt = 0;
   f = 0;
   int idx;
   int currmatidx = -1;
   bool smooth = false;
   for( CStrNode *sn = strlist.mHead; sn != NULL; sn = sn->mNext )
   {
      if( strlen( sn->mText ) == 0 )
      {
         continue;
      }
      // Grab the first token
      next = sn->mText;
      sscanf( next, "%s%n", token, &width );
      next = next + width;
      if( strcmp( token, "v" ) == 0 )
      {
         sscanf( next, "%e %e %e", &( mV[v].x ), &( mV[v].y ), &( mV[v].z ));
         ++v;
      }
      else if( strcmp( token, "vt" ) == 0 )
      {
         sscanf( next, "%e %e", &( mVt[vt].x ), &( mVt[vt].y ));
         ++v;
      }
      else if( strcmp( token, "f" ) == 0 || strcmp( token, "l" ) == 0 )
      {
         for( ;; )
         {
            int ct = sscanf( next, "%s%n", token2, &width );
            if( ct != 1 )
            {
               break;
            }
            next = next + width;
            ct = sscanf( token2, "%d%n", &idx, &width );
            if( ct != 1 )
            {
               break;
            }
            mF[f].Append( idx - 1 );
            mF[f].mMatIndex = currmatidx;
            mF[f].mSmooth = smooth;
            // Set the type to poly or line
            if( strcmp( token, "l" ) == 0 )
            {
               mF[f].mIsLine = true;
            }
         }
         ++f;
      }
      else if( strcmp( token, "s" ) == 0 )
      {
         int ct = sscanf( next, "%s%n", token2, &width );
         if( ct == 1 )
         {
            if( strcmpi( token2, "off" ) == 0 || atoi( token2 ) == 0 )
            {
               smooth = false;
            }
            else
            {
               smooth = true;
            }
         }
      }
      else if( strcmp( token, "usemtl" ) == 0 )
      {
         while( *next == ' ' )
         {
            ++next;
         }
         currmatidx = FindMaterial( next );
      }
   }

}
//---------------------------------------------------------------------------------------
