#ifndef cobj_H
#define cobj_H

#include <stdio.h>
#include <string.h>

#define COBJ_MAXSTRLEN 8192

//---------------------------------------------------------------------------------------
// Material
class CObjMat
{
public:
   // Material name
   char *mName;
   // Ka
   float mAmbient[3];
   // Kd
   float mDiffuse[3];
   // Ks
   float mSpecular[3];
   // Tr or d
   float mTransparency;
   // Ns
   float mShininess;
   // map_Ka
   char *mTexture;
   int mTag;
   CObjMat()
   {
      mAmbient[0] = mAmbient[1] = mAmbient[2] = 0.2f;
      mDiffuse[0] = mDiffuse[1] = mDiffuse[2] = 0.8f;
      mSpecular[0] = mSpecular[1] = mSpecular[2] = 1.0f;
      mTransparency = 1.0f;
      mShininess = 0.0f;
      SetTexture( "" );
      SetName( "" );
      mTag = 0;
   }
   ~CObjMat()
   {
      SetName( NULL );
      SetTexture( NULL );
   }
   // Should realy make a common function here for setting text values
   void SetTexture( char *tex )
   {
      if( mTexture != NULL )
      {
         delete [] mTexture;
         mTexture = NULL;
      }
      if( tex == NULL )
      {
         return;
      }
      size_t slen = strlen( tex );
      if( slen == 0 )
      {
         mTexture = new char[1];
         *mTexture = '\0';
      }
      else
      {
         mTexture = new char[slen + 1];
         strcpy( mTexture, tex );
      }
   }
   void SetName( char *name )
   {
      if( mName != NULL )
      {
         delete [] mName;
         mName = NULL;
      }
      if( name == NULL )
      {
         return;
      }
      size_t slen = strlen( name );
      if( slen == 0 )
      {
         mName = new char[1];
         *mName = '\0';
      }
      else
      {
         mName = new char[slen + 1];
         strcpy( mName, name );
      }
   }
};
//---------------------------------------------------------------------------------------
// Vector2
class CObjVec2
{
public:
   float x, y;
   CObjVec2()
   {
      x = y = 0;
   }
};
//---------------------------------------------------------------------------------------
// Vector3
class CObjVec3
{
public:
   float x, y, z;
   CObjVec3()
   {
      x = y = z = 0;
   }
};
//---------------------------------------------------------------------------------------
// Face vertex
class CObjFaceVert
{
public:
   CObjFaceVert *mNext;
   int mVidx;
   CObjFaceVert()
   {
      mNext = NULL;
      mVidx = -1;
   }
};
//---------------------------------------------------------------------------------------
// Face
class CObjFace
{
public:
   int mMatIndex;
   bool mIsLine;
   bool mSmooth;
   CObjFaceVert *mHead;
   CObjFaceVert *mTail;
   CObjFace()
   {
      mHead = NULL;
      mTail = NULL;
      mMatIndex = -1;
      mIsLine = false;
      mSmooth = false;
   }
   ~CObjFace()
   {
      Clear();
   }
   // Clear the list
   void Clear()
   {
      while( mHead != NULL )
      {
         mTail = mHead;
         mHead = mHead->mNext;
         delete mTail;
      }
      mTail = NULL;
   }
   void Append( int idx )
   {
      if( mHead == NULL )
      {
         mHead = new CObjFaceVert();
         mTail = mHead;
      }
      else
      {
         mTail->mNext = new CObjFaceVert();
         mTail = mTail->mNext;
      }
      mTail->mVidx = idx;
   }
};
//---------------------------------------------------------------------------------------
class CObj
{

private:

   int FindMaterial( char *matname );

public:

   CObjVec3 *mV;
   CObjVec2 *mVt;
   CObjFace *mF;
   CObjMat *mM;

   int mVCount;
   int mVtCount;
   int mFCount;
   int mMCount;

   CObj();
   ~CObj();

   // Clear obj data
   void Clear();

   // Load OBJ data from a string list
   void LoadFromFile( char *fname );

};
//---------------------------------------------------------------------------------------

#endif
