#ifndef cstrlist_H
#define cstrlist_H

#include <string.h>
#include <stdio.h>
#include "cexcept.h"

//---------------------------------------------------------------------------------------
// String node
class CStrNode
{

public:

   CStrNode *mNext;
   CStrNode *mPrev;
   char *mText;

   CStrNode()
   {
      mNext = NULL;
      mPrev = NULL;
      mText = NULL;
   }
   ~CStrNode()
   {
      ClearText();
   }

   // Clear the node's text
   void ClearText()
   {
      if( mText != NULL )
      {
         delete [] mText;
         mText = NULL;
      }
   }

   // Set the text
   void SetText( char *text )
   {
      ClearText();
      if( text == NULL )
      {
         return;
      }
      else if( *text == '\0' )
      {
         mText = new char[1];
         *mText = '\0';
      }
      else
      {
         mText = new char[strlen( text ) + 1];
         strcpy( mText, text );
      }
   }

};
//---------------------------------------------------------------------------------------
// String list
class CStrList
{

private:

public:

   // List
   CStrNode *mHead;
   CStrNode *mTail;

   CStrList()
   {
      mHead = NULL;
      mTail = NULL;
   }
   ~CStrList()
   {
      Clear();
   }

   // Clear the list
   void Clear()
   {
      while( mHead != NULL )
      {
         mTail = mHead;
         mHead = mHead->mNext;
         delete mTail;
      }
      mTail = NULL;
   }

   // Add a string to the end of the list
   void Append( char *text )
   {
      if( mHead == NULL )
      {
         mHead = new CStrNode();
         mTail = mHead;
      }
      else
      {
         mTail->mNext = new CStrNode();
         mTail->mNext->mPrev = mTail;
         mTail = mTail->mNext;
      }
      mTail->SetText( text );
   }

   // Load string list from a text file
   void LoadFromFile( char *filename, int maxchars )
   {
      FILE *fp = fopen( filename, "rt" );
      if( fp == NULL )
      {
         throw CExcept( "Error loading file \"%s\" - could not find file", filename );
      }
      char *dataline = new char[maxchars];
      while( fgets( dataline, maxchars - 1, fp ) != NULL )
      {
         // Strip newline
         int slen = ( int )strlen( dataline );
         for( int ix = slen - 1; ix > -1; --ix )
         {
            if( dataline[ix] == 10 || dataline[ix] == 13 )
            {
               dataline[ix] = '\0';
            }
            else
            {
               break;
            }
         }
         // Add the string
         Append( dataline );
      }
      delete dataline;
      fclose( fp );
   }

};
//---------------------------------------------------------------------------------------

#endif
